% -------------------------------------------------------------------------
% |            [f, xiMV, Xrec] = VRE_fault_rec(X, PCA_Struct)             |
% -------------------------------------------------------------------------
% -----------------
% | FUNCTIONALITY |
% -------------------------------------------------------------------------
% | This function aim is to first identify the fault direction affecting  |
% | each faulty observation. Then, the method computes the fault          |
% | magnitude in every variable presenting the abnormal behaviour, and    |
% | finally, reconstructing each faulty observation. To do so, the method |
% | employs the reconstruction-based contributions (Alcala and Qin, 2009) |
% | to identify the variables with an abnormal behaviour. Then, the       |
% | fault magnitude and fault reconstruction are computed using the       |
% | formulae in (Dunia and Qin, 1996).                                    |
% -------------------------------------------------------------------------
% --------------------
% | INPUT PARAMETERS |
% -------------------------------------------------------------------------
% | X:          Array or matrix of double values whose observations       |
% |             (rows) have to be reconstructed (if their square          |
% |             prediction error surpasses the statistical limit of the   |
% |             PCA model).                                               |
% | PCA_Struct:	Structure that contains the information required to       |
% |             project new observations to a given principal component   |
% |             analysis (PCA) model. This structure must have, at least: |
% |             -> mx: array containing the mean for each variable        |
% |             (column) to subtract to X.                                |
% |             -> std: array containing the standard deviation of each   |
% |             variables (column) in X. This value is used to rescalate  |
% |             (divide) each variable.                                   |
% |             -> P: matrix containing the weights (loadings) to project |
% |             each observation (row) in X to the PCA model.             |
% |             -> SPE_lim: statistical limit for the SPE. This value is  |
% |             used to determine if an observation (row) in X is faulty  |
% |             (greater than SPE_lim) or not (equal or lower).           |
% -------------------------------------------------------------------------
% ---------------------
% | OUTPUT PARAMETERS |
% -------------------------------------------------------------------------
% | f:      Array containing the fault magnitude for each faulty          |
% |         observation in X. A non-faulty observation has f set to 0,    |
% |         while any value different value indicates that the            |
% |         observation is faulty.                                        |
% | xiMV:   Array or matrix (if X was a matrix) used to reconstruct each  |
% |         observation (row) in X. Each row in xiMV is the fault         |
% |         direction that affects the corresponding observation in X. A  |
% |         non-faulty observation has all elements set to 0, while a     |
% |         faulty observation has the corresponding elements affected by |
% |         the fault set to a value different to 0. This value           |
% |         represents the unitary vector representing the projection of  |
% |         the fault for the variable (column) in question.              |
% | Xrec:   Reconstructed values for the faulty observations that         |
% |         surpassed the SPE limit. Those observations that were not     |
% |         considered as faulty have the same values, while the faulty   |
% |         ones present only variations on the variables affected by the |
% |         the corresponding fault direction.                            |
% -------------------------------------------------------------------------
function [f, xiMV, Xrec] = VRE_fault_rec(X, PCA_Struct)

%% PARAMETER CHECK

% First of all, we check if the user passed all required arguments to
% proceed with the function.
if nargin ~= 2
    % - We inform the user that we need 2 input parameters.
    error('VRE_fault_rec:notEnoughInputs', ...
        'ERROR: The method requires two input parameters!');
end

% Next, we check if the first input (X) is a numeric array or matrix
if ~isnumeric(X)
    % - We inform the user that X must be a numeric array or matrix.
    error('VRE_fault_rec:wrongType', ...
        'ERROR: X must be a numeric array or matrix!');
end

% Finally, we check if the second input parameter is a structure...
if ~isstruct(PCA_Struct)
    % - We inform the user that PCA_Struct must be a Matlab structure
    % object.
    error('VRE_fault_rec:wrongType', ...
        'ERROR: PCA_Struct must be a structure!');
else
    % - If PCA_Struct is a structure, it has to have somespecific fields:
    % -- Mean (mx):
    if ~isfield(PCA_Struct, 'mx')
        error('VRE_fault_rec:missingField', ...
            'ERROR: PCA_Struct must have the field mx!');
    end
    % -- Standard deviation (std):
    if ~isfield(PCA_Struct, 'std')
        error('VRE_fault_rec:missingField', ...
            'ERROR: PCA_Struct must have the field std!');
    end
    % -- Loading matrix (P):
    if ~isfield(PCA_Struct, 'P')
        error('VRE_fault_rec:missingField', ...
            'ERROR: PCA_Struct must have the field P!');
    end
    % -- Square prediction limit (SPE_lim):
    if ~isfield(PCA_Struct, 'SPE_lim')
        error('VRE_fault_rec:missingField', ...
            'ERROR: PCA_Struct must have the field SPE_lim!');
    end
end

%% INITIALISATIONS.

% The first step will consist on auto scaling the given data matrix (X) 
% using the mean and standard deviations obtained to build the PCA model.
X_as = standardise_to(X, PCA_Struct.mx, PCA_Struct.std);

% Next, we initialise all output parameters:
% - Fault magnitude (f):
f = zeros(size(X, 1), 1);
% - Fault direction to use when reconstructing each faulty observation
% (xiMV):
xiMV = zeros(size(X));
% - Values of the reconstructed faulty observations (Xrec):
Xrec = X;

% With the standardised data, we project it into the projection space of
% the PCA model (T):
T = X_as * PCA_Struct.P;

% We compute the projection error (E):
E = X_as - T * PCA_Struct.P';

% And the values of the squared prediction error (SPE):
SPE = sum(E .^ 2, 2);

% Next, we compare the SPE values with the statistical limit for this
% index in order to identify the faulty observations.
SPE_fd = SPE > PCA_Struct.SPE_lim;

% Next, we create the fault direction matrix. In this case, we are only
% interested in the reconstruction of individual sensor faults, therefore,
% the resulting fault direction matrix is the identity matrix.
Xi = eye(size(X, 2));

% Next, we project the fault direction matrix into the residual subspace.
Xi_rs = (eye(size(X, 2)) - PCA_Struct.P * PCA_Struct.P') * Xi;

% The last step before computing the fault magnitude of each faulty
% observation is to compute the norm of each fault direction in the
% residual subpsace. 
% - To do so, we initialise the vector to store the norm
% of each direction:
normXi_rs = zeros(1, size(Xi_rs, 2));
% - And for each fault direction...
for i = 1:length(normXi_rs)
    % -- We compute the norm of the i-th fault direction.
    normXi_rs(i) = norm(Xi_rs(:, i));
end

%% FAULT MAGNITUDE AND INIDIVIDUAL VARIABLE FAULT IDENTIFICATION.

% Using all the information we have from the previous steps (fault
% directions projected into the residual subspace, residual matrix of the
% observations and norm of each fault direction) we compute the fault
% magnitude for each faulty observation.
% - First, we initialise the matrix that will contain the values of the
% fault projected in each possible fault direction.
fMat = zeros(size(X));
% - And then, we compute the fault magnitude for each faulty observation.
fMat(SPE_fd, :) = (Xi_rs' * X_as(SPE_fd, :)')' ./ ...
    repmat(normXi_rs .^ 2, sum(SPE_fd), 1);

% We also compute the reconstruction-based contributions in order to grant
% that at least individual variable faults can be correctly diagnosed.
RBC_Struct = compute_RBC(X_as, PCA_Struct);

% Next, the first fault direction candidate is the one that presents a
% higher RBC to the SPE index.
[~, SPE_RBC_sorted] = sort(RBC_Struct.SPE.RBC, 2, 'descend');

% We reconstruct the fault directions using the fault direction associated
% to the maximum RBC value for each faulty observation.
% - First, we create the variable to store the different fault directions
% to used when reconstructing...
Xi_rec1 = zeros(size(X));
% - And now, we indicate the variables to reconstruct (only one at a time)
% for each faulty observation.
Xi_rec1(sub2ind(size(Xi_rec1), (1:size(X, 1))', SPE_RBC_sorted(:, 1))) = 1;
% - Next, we indicate the fault magnitude of the fault projected in the
% first candidate direction.
f_rec1 = fMat(sub2ind(size(fMat), (1:size(X, 1))', SPE_RBC_sorted(:, 1)));
% - And with these two values, we proceed to the reconstruction of the
% faulty observations (considering only individual variable faults).
X_rec_as1 = X_as - (Xi_rec1 .* repmat(f_rec1, 1, size(X, 2)));

% Next, we project the reconstructed faulty observations to the calibration
% model.
T_rec1 = X_rec_as1 * PCA_Struct.P;

% We compute the residual matrix for the reconstructed observations.
E_rec1 = X_rec_as1 - T_rec1 * PCA_Struct.P';

% And we compute the SPE for the reconstructed observations.
SPE_rec1 = sum(E_rec1 .^ 2, 2);

% Next, we compare the new values of the SPE with the SPE statistical
% control limit in order to find those observations whose misbehaviour has
% been corrected.
SPE_rec1_fd = SPE_rec1 > PCA_Struct.SPE_lim;

% Those observations whose SPE value after being reconstructed are below
% the statistical control limit have been correctly reconstructed.
% Therefore, we can copy these values to the corresponding variables.
% - First, we locate the positions to overwrite as those that were
% initially over the SPE limit and once reconstructed are below the SPE
% limit.
overwrite = xor(SPE_fd, SPE_rec1_fd);
% - Now, we desnormalise the values of the reconstructed faulty
% observations.
Xrec(overwrite, :) = X_rec_as1(overwrite, :) .* ...
    repmat(PCA_Struct.std, sum(overwrite), 1) + ...
    repmat(PCA_Struct.mx, sum(overwrite), 1);

% We also update the values of the fault magnitude to reconstruct those
% positions:
f(overwrite) = f_rec1(overwrite);

% And the fault direction to reconstruct those observations.
xiMV(overwrite, :) = Xi_rec1(overwrite, :);

%% MULTIVARIABLE FAULT TREATEMENT.

% In this part of the code we will deal with multivariable faults ocurring
% on the studied process. In the previous section we reconstructed the
% individual variable faults, so in this section we will focus on those
% observations for which the SPE value surpasses the SPE limit even after
% reconstructing one of the variables.

% The first step is to know how faults propagate from one variable to the
% rest in the residual space (since we are detecting faults using the SPE).
% Initially, we compute this value considering a unit fault, and then we
% will compute the expected value considering the fault magnitude estimated
% in the initial fault direction reconstruction.
% - The unit RBC propagation of a fault depends solely on the relation
% between the residual covariance between each pair of variables and its
% own residual variance. So first, we compute the residual covariance
% matrix.
Crs = eye(size(X, 2)) - PCA_Struct.P * PCA_Struct.P';
% - Now, we can compute the RBC propagation of a unit fault magnitude for
% each pair of variables.
cProp = (Crs .^ 2) ./ repmat(diag(Crs)', size(X, 2), 1);
% - Next, we compute how we had expected the contributions of the faulty
% observation should an individual sensor fault had happened.
expectedProp = cProp(SPE_RBC_sorted(:, 1), :) .* ...
    (repmat(fMat(sub2ind(size(fMat), (1:size(fMat, 1))', ...
    SPE_RBC_sorted(:, 1))), 1, size(cProp, 2)) .^ 2);
% - And now, we compute the difference between the expected value and the
% real value we've found.
difProp = RBC_Struct.SPE.RBC - expectedProp;
% - Finally, we sort the differences from higher (the next candidate) to 
% lower (the one we've reconstructed).
[~, mvCandidatesSorted] = sort(difProp, 2, 'descend');

% Now comes the iterative part to identify the fault affecting each
% observation.
% - First, we initialise the variable that will control the loop.
loop = sum(SPE_rec1_fd) > 0 ;
% - Next, we initialise the variable that indicates how many variables have
% to be used to reconstruct each faulty observation...
up2 = 1;
% - We also create a variable to store the previous values of the SPE in
% order to determine which observations can be reconstructed in the current
% iteration.
SPE_ant_fd = SPE_rec1_fd;
% - We also store the values of the previously reconstructed values (in the
% auto scaled space).
X_rec_as_ant = X_rec_as1;
% - While we have to loop...
while loop
    % -- We initialise the fault direction matrix to reconstruct the faulty
    % observations.
    Xi_rec_i = zeros(size(X));
    
    % -- Next, we form the list of the candidates to use when
    % reconstructing each faulty observation. This uses the first candidate
    % and the set of variables with the highest difference between the
    % expected value and the real RBC.
    % --- To do so, first, we copy the fault magnitude values used for the
    % first reconstruction.
    Xi_rec_i(sub2ind(size(Xi_rec_i), find(SPE_ant_fd), ...
        SPE_RBC_sorted(SPE_ant_fd, 1))) = ...
        fMat(sub2ind(size(Xi_rec_i), find(SPE_ant_fd), ...
        SPE_RBC_sorted(SPE_ant_fd, 1)));
    % --- Next, we find the positions of the faulty observations.
    fPos_i = find(SPE_ant_fd);
    % --- Now, for each faulty observation...
    for i = 1:length(fPos_i)
        % ---- We copy all the values of the fault direction candidates for
        % the i-th faulty observation.
        Xi_rec_i(fPos_i(i), mvCandidatesSorted(fPos_i(i), 1:up2)) = ...
            fMat(fPos_i(i), mvCandidatesSorted(fPos_i(i), 1:up2));
    end
%     % --- Next, we add the rest of reconstruction candidates, whose
%     % difference between the expected value and the current RBC values.
%     Xi_rec_i(sub2ind(size(Xi_rec_i), find(SPE_ant_fd), ...
%         mvCandidatesSorted(SPE_ant_fd, 1:up2))) = ...
%         fMat(sub2ind(size(Xi_rec_i), find(SPE_ant_fd), ...
%         mvCandidatesSorted(SPE_ant_fd, 1:up2)));
        
    % -- We compute the norm of each row, which is equivalent to computing
    % the global fault magnitude affecting the observations.
    % --- First of all, we initialise the variable to ones, since we will
    % be dividing all values by this value.
    f_i = ones(size(X, 1), 1);
    % --- Next, we compute the norm of all faulty observations in the
    % previous iteration.
    f_i(SPE_ant_fd) = sqrt(sum(Xi_rec_i(SPE_ant_fd, :) .^ 2, 2));
      
    % -- Finally, we divide each observation's fault direction by the fault
    % magnitude affecting the whole process.
    Xi_rec_i = Xi_rec_i ./ repmat(f_i, 1, size(Xi_rec_i, 2));
    
    % -- Next, we proceed to reconstruct the faulty observations using both
    % the multivariable fault directions and the global fault magnitude.
    % --- First, we reconstructed values in the previous observation...
    X_rec_as_i = X_rec_as_ant;
    % --- Now, we reconstruct the multivariate faulty observations.
    X_rec_as_i(SPE_ant_fd, :) = X_rec_as_i(SPE_ant_fd, :) - ...
        (Xi_rec_i(SPE_ant_fd, :) .* repmat(f_i(SPE_ant_fd), 1, ...
        size(X, 2)));
    % --- We project the reconstructed observations into the PCA model.
    T_i = X_rec_as_i * PCA_Struct.P;
    % --- We also compute the residual matrix of the observations.
    E_i = X_rec_as_i - T_i * PCA_Struct.P';
    % --- And the SPE values.
    SPE_i = sum(E_i .^ 2, 2);
    % --- Which observations are still faulty?
    SPE_i_fd = SPE_i > PCA_Struct.SPE_lim;
    
    % -- Next, we look for which observations have been correctly
    % reconstructed in this iteration.
    overwrite_i = xor(SPE_ant_fd, SPE_i_fd);
    
    % -- And we update the values of both the fault magnitude and fault
    % direction for the correctly reconstructed observations in this loop.
    % --- Fault magnitude.
    f(overwrite_i) = f_i(overwrite_i);
    % --- Fault direction to use when reconstructing.
    xiMV(overwrite_i, :) = Xi_rec_i(overwrite_i, :);
    % --- Reconstructed observations denormalised.
    Xrec(overwrite_i, :) = X_rec_as_i(overwrite_i, :) .* ...
        repmat(PCA_Struct.std, sum(overwrite_i), 1) + ...
        repmat(PCA_Struct.mx, sum(overwrite_i), 1);
    
    % -- Finally, we check if we have to loop (there are still some 
    % observations that have to be reconstructed) or not (all faulty
    % observations were correctly reconstructed), or the observations
    % cannot be reconstructed (all variables used to reconstruct).
    loop = sum(SPE_i_fd) > 0 && up2 < (size(xiMV, 2) - 1);
    
    % -- If we have to loop another time...
    if loop
        % --- Then, we have to update some variables for the next
        % iteration.
        % ---- Position to use for the next reconstruction step.
        up2 = up2 + 1;
        % ---- Faulty observations found on the previous iteration after
        % reconstruction.
        SPE_ant_fd = SPE_i_fd;
        % ---- Previously reconstructed values (on the auto scaled space).
        X_rec_as_ant = X_rec_as_i;
    end
end