% -------------------------------------------------------------------------
% |              RBC_ind_struct = compute_RBC(X, PCA_struct)              |
% -------------------------------------------------------------------------
% -----------------
% | FUNCTIONALITY |
% -------------------------------------------------------------------------
% | This function aim is to compute the reconstruction-based              |
% | contributions (RBC) to an observation dataset (X). The RBC is         |
% | computed for the squarred prediction error (SPE), the Hottelling's    |
% | T^2 statistic and the  combined index Phi.                            |
% -------------------------------------------------------------------------
% --------------------
% | INPUT PARAMETERS |
% -------------------------------------------------------------------------
% | X:          Dataset containing the observations to compute the RBC    |
% |             for the three different indices. X can be either an auto  |
% |             scaled  vector or a matrix of real values.                |
% | PCA_struct:	Structure that contains the information of a PCA model.   |
% |             This structure must have the following fields, at least:  |
% |             -> P. The loading matrix used to project new observations |
% |             into the PCA model.                                       |
% |             -> lambda. Row vector containing the eigenvalues          |
% |             associated to the retained principal components.          |
% |             -> SPE_lim. Limit for the SPE index.                      |
% |             -> T2_lim. Limit for the Hottelling's T^2 statistic.      |
% -------------------------------------------------------------------------
% ---------------------
% | OUTPUT PARAMETERS |
% -------------------------------------------------------------------------
% | RBC_ind_struct:	Structure containing the RBC index for the            |
% |                 observation's dataset. This structure contains the    |
% |                 following information for each index (SPE, T^2 and    |
% |                 Phi, which are the main fields of the structure):     |
% |                 -> RBC. Matrix containing the reconstruction-based    |
% |                 contribution for each variable (column) and           |
% |                 observation (row).                                    |
% |                 -> RBC_fd. Logical (with only 0's and 1's) matrix     |
% |                 that indicates which variables surpass the RBC        |
% |                 control limit. Rows of this matrix refer to           |
% |                 observations, while columns are the different         |
% |                 variables.                                            |
% |                 -> RBC_limit. Row vector containing the RBC control   |
% |                 limit for each variable.                              |
% -------------------------------------------------------------------------

function RBC_ind_struct = compute_RBC(X, PCA_struct)
%% INITIALISATIONS.
% In this step we will initialise the different variables to use throughout
% all the function.
% Firstly, we create the variable that we will use to get the RBC for each
% variable (an identity matrix with as many rows and columns as variables
% the observation matrix has).
Xi = eye(size(X, 2));
% Next, we compute the different variables required to compute the RBC
% values for each of the three indices (SPE, T^2 and Phi).
% - Related with the SPE index.
% -- Residual covariance matrix.
C_res = Xi - PCA_struct.P * PCA_struct.P';
% - Related with the T^2 index.
% -- Diagonal matrix containing the eigenvalues associated to the retained
% principal components inverted (each element in the diagonal is equal 1 / 
% sqrt(lambda(i)).
Lambda_Inv = diag(1 ./ sqrt(PCA_struct.lambda));
% -- Since Lambda contains zeros except in the diagonal, we can't invert it
% directly. In this case, we have to invert only the elements of the
% diagonal. This is equivalent to create a diagonal matrix whose values are
% 1 / sqrt(Lambda)
% -- Matrix containing the rescaled loading vectors.
D = PCA_struct.P * Lambda_Inv * PCA_struct.P';
% - Related with the combined index Phi.
% -- Matrix to compute the combined index Phi for new observations.
Phi = C_res / PCA_struct.SPE_lim + D / PCA_struct.T2_lim;
% Now, we initialise a set of variables associated to the computation of
% each RBC index control limit and values.
% - Related to the SPE.
% -- Amplitude of the chi-square distribution to compute the RBC control
% limit.
amplitude_SPE = zeros(1, size(Xi, 2));
% -- Matrix to store the RBC of the SPE index.
RBC_SPE = zeros(size(X));
% - Related to the T^2.
% -- Amplitude of the chi-square distribution to compute the RBC control
% limit.
amplitude_T2 = zeros(1, size(Xi, 2));
% -- Matrix to store the RBC of the SPE index.
RBC_T2 = zeros(size(X));
% - Related to the combined index Phi.
% -- Amplitude of the chi-square distribution to compute the RBC control
% limit.
amplitude_Phi = zeros(1, size(Xi, 2));
% -- Matrix to store the RBC of the SPE index.
RBC_Phi = zeros(size(X));

%% COMPUTING THE RBC FOR EACH INDEX.
% In this cell, the RBC information for each index is computed and added to
% the RBC structure.

% Firstly, we will compute the reconstruction-based contributions limits
% for each index.
% - For each fault direction...
for i = 1:size(Xi, 2)
    % -- Firstly, we compute the amplitude of the RBC for each index.
    % --- SPE index.
    amplitude_SPE(i) = Xi(:, i)' * C_res * PCA_struct.R * ...
        C_res * Xi(:, i) / (Xi(:, i)' * C_res * Xi(:, i));
    % --- T^2 index.
    amplitude_T2(i) = Xi(:, i)' * D * PCA_struct.R * D * Xi(:, i) / ...
        (Xi(:, i)' * D * Xi(:, i));
    % --- Phi index.
    amplitude_Phi(i) = Xi(:, i)' * Phi * PCA_struct.R * Phi * ...
        Xi(:, i) / (Xi(:, i)' * Phi * Xi(:, i));
    % -- At the same time, we compute the RBC to the i-th variable.
    % --- SPE index.
    RBC_SPE(:, i) = (Xi(:, i)' * C_res * X') .^ 2 / C_res(i,i);
    % --- T^2 index.
    RBC_T2(:, i) = (Xi(:, i)' * D * X') .^ 2 / D(i,i);
    % --- Phi index.
    RBC_Phi(:, i) = (Xi(:, i)' * Phi * X') .^ 2 / Phi(i,i);
end
% Now that we have the amplitudes of the chi-square distribution, we can
% compute the control limits for the RBC of each index.
% - SPE index.
RBC_SPE_limit = amplitude_SPE * chi2inv(0.95, 1);
% - T^2 index.
RBC_T2_limit = amplitude_T2 * chi2inv(0.95, 1);
% - Phi index.
RBC_Phi_limit = amplitude_Phi * chi2inv(0.95, 1);

% Next, using the control limits we have just found, we determine the
% sensors that surpass the control limit for each observation.
% - SPE index.
RBC_SPE_fd = RBC_SPE > repmat(RBC_SPE_limit, size(RBC_SPE, 1), 1);
% - T^2 index.
RBC_T2_fd = RBC_T2 > repmat(RBC_T2_limit, size(RBC_T2, 1), 1);
% - Combined index Phi.
RBC_Phi_fd = RBC_Phi > repmat(RBC_Phi_limit, size(RBC_Phi, 1), 1);

%% STRUCTURE BUILDING.
% Now that we have all the information, we can create the structure to
% store the information related to the reconstruction-based contributions.
% - Information related to the SPE index:
% -- RBC control limit.
RBC_ind_struct.SPE.RBC_limit = RBC_SPE_limit;
% -- RBC values.
RBC_ind_struct.SPE.RBC = RBC_SPE;
% -- Variables whose RBC values are over the control limits.
RBC_ind_struct.SPE.RBC_fd = RBC_SPE_fd;
% - Information related to the T^2 index:
% -- RBC control limit.
RBC_ind_struct.T2.RBC_limit = RBC_T2_limit;
% -- RBC values.
RBC_ind_struct.T2.RBC = RBC_T2;
% -- Variables whose RBC values are over the control limits.
RBC_ind_struct.T2.RBC_fd = RBC_T2_fd;
% - Information related to the combined index Phi:
% -- RBC control limit.
RBC_ind_struct.Phi.RBC_limit = RBC_Phi_limit;
% -- RBC values.
RBC_ind_struct.Phi.RBC = RBC_Phi;
% -- Variables whose RBC values are over the control limits.
RBC_ind_struct.Phi.RBC_fd = RBC_Phi_fd;