%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% extended_cm = compute_extended_confusion_matrix(predicted_class,       %
%%                 real_class, n_classes)                                 %
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% FUNCTIONALITY: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This function aim is to compute the extended confusion matrix for a %%%%
%% vector of predicted classes (predicted_class) by comparing them to the %
%% corresponding real class (real_class). %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% INPUT ARGUMENTS: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% predicted_class: Vector containing the predicted class for a set of %%%%
%%                  observations. This variable must a vector of integer %%
%%                  values to facilitate the confusion matrix %%%%%%%%%%%%%
%%                  computation. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% real_class:      Vector containing the real class associated to the %%%%
%%                  same observations that were predicted and passed as %%%
%%                  an input argument (predicted_class). %%%%%%%%%%%%%%%%%%
%% n_classes:       Integer value that reprsents the number of different %%
%%                  classes present in the data set. Classes must be %%%%%%
%%                  represented by integer values in range of 1 to %%%%%%%%
%%                  n_classes. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% OUTPUT ARGUMENTS: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% extende_cm:      Matrix containing the extended matrix for the set of %%
%%                  observations passed as an input argument. Its %%%%%%%%%
%%                  dimensions are n_classes times n_classes. %%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function extended_cm = compute_extended_confusion_matrix(...
    predicted_class, real_class, n_classes)
% First of all, we initialise the varibale that will contain the extended
% confusion matrix.
extended_cm = zeros(n_classes);
% Now, for each different class...
for i = 1:n_classes
    % - Firstly, we create a vector that will contain all the different
    % values of the classes.
    class_vector = 1:n_classes;
    % - We remove the class we are looking for in this iteration.
    class_vector(i) = [];
    % - We find the positions of the cases whose real class is i.
    pos_obs_class_i = real_class == i;
    % - Now, we get the predicted class for those cases.
    pred_class_obs_class_i = predicted_class(pos_obs_class_i);
    % - Now, we compute the number of true positives and false positives
    % for the i-th class.
    % -- True positives.
    extended_cm(i,i) = sum(real_class(pos_obs_class_i) == pred_class_obs_class_i);
    % -- False positives. To compute this values, we will need to loop for
    % all the other classes.
    for j = 1:length(class_vector)
        % --- We seek for the cases misclassified for the j-th class.
        extended_cm(i,class_vector(j)) = sum(pred_class_obs_class_i == class_vector(j));
    end
end