%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% [U,xi_pcs,xi_res] = compute_reconstruction_error_matrix(X,xi) %%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% FUNCTIONALITY: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This function purpose is to compute the reconstruction error matrix %%%%
%% for a given data set X for a certain fault set xi. %%%%%%%%%%%%%%%%%%%%%
%% INPUT PARAMETERS: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% X:	Scaled data set used to compute the reconstruction error matrix. %%
%% xi:	Fault data set used to compute the reconstruction error matrix. %%%
%% OUTPUT PARAMETERS: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% U:       Unreconstructed variance matrix for the data set X and the %%%%
%%          fault set xi. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% xi_pcs:  Projection of the fault matrix xi in the principal components %
%%          subspace. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% xi_res:  Projection of the fault matrix xi in the residual subspace. %%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [U,xi_pcs,xi_res] = compute_reconstruction_error_matrix(X,xi)
% We initialise the values of the reconstruction error vector (1 value for
% each possible number of principal components)
U = zeros(1,size(X,2));
% We compute the correlation matrix of the data set X
corr_X = (X' * X) / (size(X,1) - 1);
% We decompose the correlation matrix in eigenvectors and eigenvalues
[eigenvectors_X, eigenvalues_X] = eig(corr_X);
% Next, we sort in descending order the eigenvalues to obtain the PCA
% decomposition
[~, indices_sort_eigenvectors_X] = ...
    sort(diag(eigenvalues_X), 'descend');
% We sort the same way the eigenvectors, to obtain the loading vectors of
% the PCA decomposition
P = eigenvectors_X(:, indices_sort_eigenvectors_X);
% We initialise the structure that will contain all the projections in the
% residual and principal components subspaces for each number of principal
% components
% - Projection of fault directions in the principal components subspace
xi_pcs = cell(1,size(xi,2));
% - Projection of fault direction in the residual subspace
xi_res = cell(1,size(xi,2));
% We obtain the score matrix (T) and the loading vectors (P) of the pca
% model
T = X * P;
% Next, we compute the covariance matrix for the observation matrix X.
R = (X' * X) / (size(X, 1) - 1);
% For every possible number of principal components
for i = 1:size(T,2)
    % We compute the projection matrix of theta in the projection space
    xi_pcs{i} = (xi'*P(:,1:i)*P(:,1:i)')';
    % We compute the projection matrix of theta in the residual space
    xi_res{i} = xi - xi_pcs{i};
    % - Now, we can compute the variance of the reconstruction error for
    % the residual subspace projection of the faulty situations into a PCA
    % model which retains i principal components.
    % SIMPLIFIED METHOD TO COMPUTE THE VRE!
    U_i = diag(xi_res{i}' * R * xi_res{i}) ./ diag(xi_res{i}' * ...
        xi_res{i}).^2;
    % We initialise the weighting vector q (we assume all faults have the
    % same importance, so we set all of them to 1)
    q = ones(size(U_i,1),1);
    % We normalise the vector so its norm is 1
    q = q / norm(q);
    % We compute the reconstruction error vector with the i first principal components with the weighted sum
    % of the projection error and the residual error
    U(i) = q'*U_i;
end