%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% data_standardised = standardise_to(data, mean_to_scale, std_to_scale) %%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% FUNCTIONALITY: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This function aim is to standardise a data set (X) using the mean %%%%%%
%% (mean_to_scale) and the standard deviation (std_to_scale) provided by %%
%% the user. This method is intended for bidimensional data.%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% INPUT PARAMETERS: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% data:            Data set to standardise. It can be either a vector or %
%%                  a matrix. The number of columns in data must be the %%%
%%                  same as the number of elements in the mean and %%%%%%%%
%%                  standard deviation values given. %%%%%%%%%%%%%%%%%%%%%%
%% mean_to_scale:   Vector that contains the mean values for each %%%%%%%%%
%%                  variable (columns in data). The number of elements in %
%%                  this vector must be the same as the number of columns %
%%                  in data. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% std_to_scale:    Vector that contains the standard deviations to use %%%
%%                  to scale the data set (data). The number of elements %%
%%                  in this vector must be the same that the number of %%%%
%%                  columns in the data set (data). %%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% OUTPUT PARAMETERS: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% data_standardised:   Data set (data) after standardised. It has the %%%%
%%                      same dimensions that data (vector or matrix). If %%
%%                      any problem occurred while standardising data, %%%%
%%                      the returned value is NaN. Otherwise, the %%%%%%%%%
%%                      returned values are doubles. %%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function data_standardised = standardise_to(data, mean_to_scale, std_to_scale)
%% INITIALISATIONS:
% First of all we have to preallocate the memory required to standardise
% the data. To do so, we have to check if the dimensions of the mean and
% standard deviation vectors are equal to the number of columns in data.
if size(data,2) == length(mean_to_scale) && size(data,2) == length(std_to_scale)
    % - All variables have the same number of variables, so we can proceed
    % normally.
    
    % -- Now, we standardise the data using the mean and vectors the user
    % passed as input parameters.
    data_standardised = (data - repmat(mean_to_scale, size(data, 1), 1)) ./ repmat(std_to_scale, size(data, 1), 1);
else
    % - Some elements do not have the same dimensions. As a consequence, we
    % will return an error by initialising the data to standardise as a
    % NaN.
    data_standardised = nan();
end