% -------------------------------------------------------------------------
% |      [vc, vcJ, vcK, h] = variance_captured(Z, P, J, K, vars, rep)     |
% -------------------------------------------------------------------------
% -----------------
% | FUNCTIONALITY |
% -------------------------------------------------------------------------
% | This function aim is to compute the variance captured (vc) in a PCA   |
% | model depending on the number of principal components retained.       |
% | Additionally, the method also computes the variance captured for each |
% | variable (vcJ) and time sample (vcK) according to the number of       |
% | principal components retained if the input data matrix is a           |
% | batch-wise unfolded matrix (the number of columns of the pre-treated  |
% | data matrix - Z - is equal to J x K). Otherwise these values are set  |
% | to the empty set. Finally, and if the user indicated this (rep =      |
% | true), the method represents each of the computed variances captured  |
% | by the PCA model.                                                     |
% -------------------------------------------------------------------------
% --------------------
% | INPUT PARAMETERS |
% -------------------------------------------------------------------------
% | Z:      Double matrix with the pre-processed data. Rows are related   |
% |         to observations, while columns are the measured variables. In |
% |         case of batch-wise unfolded matrices, columns can be          |
% |         decomposed as variables measured along time, and thus, they   |
% |         are equal to J x K. Additionally, rows for a batch-wise       |
% |         unfolded matrix are the different batches monitored.          |
% | P:      Double matrix storing the loading matrix obtained from the    |
% |         eigenvalue decomposition of the covariance/correlation matrix |
% |         of the data matrix. Rows are the measured variables, while    |
% |         columns are the principal components retained to build the    |
% |         PCA model.                                                    |
% | J:      Integer value that indicates the number of variable mesured   |
% |         for each observation. Must be a value greater than 1.         |
% | K:      Integer value that indicates the number of samples measured   |
% |         for each variable. Must be a value greater than 0. When K > 1 |
% |         we are using unfolded data matrices. This means that the      |
% |         number of columns in Z must be equal to J x K.                |
% | vars:   Cell array that contains the labels of the measured           |
% |         variables. It must have J elements.                           |
% | rep:    Boolean that indicates if the variance captured computed (vc, |
% |         (and vcJ and vcK when applicable) has to be represented (rep  |
% |         = true) or not (rep = false).                                 |
% -------------------------------------------------------------------------
% ---------------------
% | OUTPUT PARAMETERS |
% -------------------------------------------------------------------------
% | vc:     Double matrix indicating the percentage of variance captured  |
% |         for each variable measured (columns in Z). It has as many     |
% |         rows as principal components were retained when building the  |
% |         PCA model (size(P, 2) and as many columns as variables        |
% |         measured (size(Z, 2)). Values returned in this matrix are in  |
% |         the following range: [0, 100]%.                               |
% | vcJ:    Double matrix containing the percentage of variance explained |
% |         for each measured variable. These values are only computed    |
% |         when the matrix comes from a batch-wise unfolded matrix       |
% |         (size(Z, 2) == J x K), and otherwise the returned value is    |
% |         the empty set. Values in this matrix are in range [0, 100]%.  |
% | vcK:    Double matrix containing the percentage of variance explained |
% |         for each sample measured. These values are computed only when |
% |         the data matrix (Z) comes from a batch-wise unfolded matrix   |
% |         (size(Z, 2) == J x K). Otherwise, the value returned is the   |
% |         empty set. Values in this matrix are in range [0, 100]%.      |
% | h:      Array of double values (in case of a batch-wise unfolded      |
% |         matrix) or double value (otherwise) indicating the handles to |
% |         the representations of the variance captured for vc, vcJ and  |
% |         vcK (when applicable).                                        |
% -------------------------------------------------------------------------

function [vc, vcJ, vcK, h] = variance_captured(Z, P, J, K, vars, rep)

%% PARAMETER CHECKING

% First of all, we check if the user passed all input parameters.
if nargin < 6
    % No, we are missing some parameters, so we inform of the problem.
    error('variance_captured:paramCheck', ...
        'ERROR: Method needs 5 input parameters!');
end

% Next, we check if the input parameters are correct.

% - Z -
% Is it a numeric matrix?
if ~isnumeric(Z)
    % - No, so we inform of the problem.
    error('variance_captured:wrongType', ...
        'ERROR: Z must be a numeric type!');
else
    % - We get the dimensions of Z.
    dims = size(Z);
    % - Is it a two-dimensional matrix?
    if length(dims) ~= 2
        % -- No, so we inform the user of the problem.
        error('variance_captured:wrongType', ...
            'ERROR: Z must be a two-dimensional matrix!');
    else
        % Is it a matrix?
        if dims(1) <= 1 || dims(2) <= 1
            % -- No, so we inform the user of the problem.
            error('variance_captured:wrongType', ...
                'ERROR: Z must be a two-dimensional matrix');
        end
    end
end

% - P -
% Is it a numeric matrix?
if ~isnumeric(P)
    % - No, so we inform of the problem.
    error('variance_captured:wrongType', ...
        'ERROR: P must be a numeric type!');
else
    % - We get the dimensions of Z.
    dims = size(P);
    % - Is it a two-dimensional matrix?
    if length(dims) ~= 2
        % -- No, so we inform the user of the problem.
        error('variance_captured:wrongType', ...
            'ERROR: P must be a two-dimensional matrix!');
    else
        % Is it a matrix?
        if dims(1) <= 1 || dims(2) <= 1
            % -- No, so we inform the user of the problem.
            error('variance_captured:wrongType', ...
                'ERROR: Z must be a two-dimensional matrix');
        elseif dims(1) ~= size(Z, 2)
            % -- It does not have the same number of measured variables, so
            % we inform the user of this problem.
            error('variance_captured:wrongDimensions', ...
                'ERROR: Rows in P must be equal to columns in Z!');
        end
    end
end

% - J - 
% Is it a numeric value?
if ~isnumeric(J)
    error('variance_captured:wrongType', ...
        'ERROR: J must be a numeric value!');
end

% Is it a value?
if size(J, 1) ~= 1 || size(J, 2) ~= 1
    % - No, it isn't. We inform the user of the problem.
    error('variance_captured:wrongType', ...
        'ERROR: J must be a [1 x 1] value!');
end

% Is within the range?
if J <= 1
    % - No, it isn't. We inform the user of the problem.
    error('variance_captured:outOfRange' , ...
        'ERROR: J must be a value greater than 1!');
end

% - K -
% Is it a numeric value?
if ~isnumeric(K)
    error('variance_captured:wrongType', ...
        'ERROR: K must be a numeric value!');
end

% Is it a value?
if size(K, 1) ~= 1 || size(K, 2) ~= 1
    % - No, it isn't. We inform the user of the problem.
    error('variance_captured:wrongType', ...
        'ERROR: K must be a [1 x 1] value!');
end

% Is within the range?
if K <= 0
    % - No, it isn't. We inform the user of the problem.
    error('variance_captured:outOfRange' , ...
        'ERROR: K must be a value greater than 0!');
elseif K == 1
    % - Additionally, if K is equal to 1, we are working with a
    % non batch-wise unfolded data matrix, so we have that size(Z, 2) is
    % equal to J.
    if J ~= size(Z, 2)
        % -- It's not the case, so we inform of the problem.
        error('variance_captured:valueNotValid', ...
            ['ERROR: When K == 1, J must be equal to the number of ', ...
            'columns in Z!']);
    end
    % - We indicate that we are not dealing with batch-wise unfolded data
    % matrices.
    bw = false;
    % - And thus, we do not have to compute the other variance captured
    % values (vcJ and vcK).
    vcJ = []; vcK = [];
else
    % - We are working with batch-wise unfolded matrices, so we check if J
    % x K is equal to the number of columns in Z.
    if J * K ~= size(Z, 2)
        % -- It's not the case, so we inform of the problem.
        error('variance_captured:valueNotValid', ...
            ['ERROR: When K > 1, J x K must be equal to the number ', ...
            'of columns in Z!']);
    end
    % - Finally, we indicate that we are working with batch-wise unfolded
    % data matrices.
    bw = true;
end

% - vars -
% Is it a cell array of strings?
if ~iscellstr(vars)
    % - No, it isn't, so we inform of the problem.
    error('variance_captured:wrongType', ...
        'ERROR: vars must be a cell array of strings!');
elseif length(vars) ~= J
    % - vars does not have the same number of measured variables, so we
    % inform of the situation.
    error('variance_captured:wrongDimensions', ...
        'ERROR: vars must have J elements!');
end
    
% - rep -
% Is it a boolean variable?
if ~islogical(rep)
    % - No, it isn't. Therefore, we indicate the problem.
    error('variance_captured:wrongType', ...
        'ERROR: rep must be a boolean variable!');
end

%% COMPUTATION OF THE VARIANCE CAPTURE FOR EACH COLUMN IN Z.

% First of all, we compute the total variation of the original data.
tV = sum(Z .^ 2);

% Now, we define the variable to store the normalised loading values.
nP = zeros(size(P));

% And we initialise the variable to store the variance capture for each
% variable and principal component.
vc = zeros(size(P, 2), size(Z, 2));

% Now, for each principal component retained...
for i = 1:size(P, 2)
    % - We compute the normalised loading vector...
    nP(:, i) = P(:, i) / norm(P(:, i));
    % - Now, we compute the estimated X matrix for the i-th loading vector.
    Zhat = Z * nP(:, i) * nP(:, i)';
    % - And finally, we compute the variance captured for each variable
    % using the i-th principal component.
    vc(i, :) = 100 * sum(Zhat .^ 2) ./ tV;
end

% Do we have to represent the variance captured?
if rep
    % - Yes, we have. But, how many handle figures we need? If we are not
    % dealing with batch-wise unfolded matrices (K == 1), we only need 1
    % handle. Ohterwise, we need 3 of them (one for vc, another to
    % represent vcJ and the last one to represent the vcK).
    if bw
        % - We need three handles, so we initialise the variable to store
        % the handles.
        h = zeros(1, 3);
        % - And the first one is the handle associated with the variance
        % captured for each column in Z.
        h(1) = figure;
    else
        % - We only need one handle, so we initialise its value now.
        h = figure;
    end
    % Now, we represent the stacked variance captured for each principal
    % component.
    bar(vc', 'Stacked');
    % Now, we redimensionate axis.
    axis([0, size(Z, 2) + 1, 0, max(sum(vc)) * 1.1]);
    % If we have a batch-wise unfolded matrix...
    if bw
        % For each variable (minus 1)...
        for i = 1 : J - 1
            % We add a line to divide variable i from variable i + 1.
            line([i * K, i * K], [0, max(sum(vc)) * 1.1], ...
                'LineStyle', '-', 'Color', 'k');
        end
        % And now, we label the figure axes.
        xlabel('Variables along time');
        title('Variance captured for each variable and sample');
    else
        % Since we do not have variables along time, we label differently
        % the axes.
        xlabel('Measured variable');
        title('Variance captured for each measured variable');
    end
    % And finally, we indicate the variable described within each division.
    set(gca, 'XTick', K / 2 : K : size(Z, 2) * K, ...
        'XTickLabel', vars);
    % We label the figure axes (the ones remainig).
    ylabel('Variance captured');
end

%% VARIANCE CAPTURE FOR EACH VARIABLE AND EACH TIME SAMPLE.

% In this stage, we compute the variance capture for each variable
% (accumulating all time samples for that variable for each principal
% component retained) and each time sample (we accumulate the variance
% captured for each variable at a given time sample for each principal
% component retained), but only if we are dealing with batch-wise unfolded
% data matrices.
if bw
    % First, we initialise the variable to store the variance captured for
    % each variable for each principal component retained.
    vcJ = zeros(size(P, 2), J);
    
    % Now, for each measured variable...
    for i = 1:J
        % - We compute the total variance for that variable for all
        % retained principal components.
        vJi = sum(sum(vc(:, (i - 1) * K + 1 : i * K)));
        % - Now, for each principal component retained...
        for j = 1:size(P, 2)
            % -- We compute the variance captured for all samples of the
            % i-th variable based on the j-th principal component.
            vcJ(j, i) = sum(vc(j, (i - 1) * K + 1 : i * K)) * 100 / vJi;
        end
    end
    
    % Now, that we have the information, do we have to represent it?
    if rep
        % - Yes, so we create the figure to represent the information.
        h(2) = figure;
        % - Now, we represent the stacked variance captured.
        bar(vcJ', 'Stacked');
        % - Next, we redimensionate the figure axes.
        axis([0, J + 1, 0, max(sum(vcJ)) * 1.1]);
        % - And finally, we indicate to which variable we are explaining
        % the variance captured.
        set(gca, 'XTick', 1 : J, 'XTickLabel', vars);
        % - And we label the figure axes.
        xlabel('Variable measured'); ylabel('Variance captured');
        title('Variance captured for each variable (samples accumulated)');
    end
    
    % The last step is to compute the variance captured for each time
    % sample (we accumulate the variance captured for all variables at that
    % time sample). First, we create the variable to store the values.
    vcK = zeros(size(P, 2), K);
    
    % Now, for each sample...
    for i = 1 : K
        % - We compute the total variance captured for the i-th time sample
        % for all principal components and variables.
        vKi = sum(sum(vc(:, i : K : size(Z, 2))));
        % - And now, for each principal component retained...
        for j = 1 : size(P, 2)
            % -- We compute the variance captured for all variables for the
            % i-th sample for the j-th principal component retained.
            vcK(j, i) = sum(vc(j, i : K : size(Z, 2))) * 100 / vKi;
        end
    end
    
    % Now, that we have the information, do we have to represent it?
    if rep
        % - Yes, so we update the handle's array with the new figure.
        h(3) = figure;
        % - Next, we represent the bar plot.
        bar(vcK', 'Stacked');
        % - Now, we redimensionate the figure axis.
        axis([0, K + 1, 0, max(sum(vcK)) * 1.1]);
        % - And we label the figure axes.
        xlabel('Time sample'); ylabel('Variance captured');
        title('Variance captured for each sample (variables accumlated)');
    end
    
end