%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% VRE_Struct = variance_reconstruction_error(X, xi) %%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% FUNCTIONALITY: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This function aim is to find the optimal number of principal %%%%%%%%%%%
%% components to grant the minimum reconstruction error for a given set %%%
%% of faults. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% INPUT PARAMETERS: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% X:	Scaled data set (zero mean and unit variance) to find its minimum %
%%      reconstruction error. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% xi:	Initial set of faults to find its minimum reconstruction error. %%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% OUTPUT PARAMETERS: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% VRE_Struct:  Structure that contains all output elements obtained %%%%%%
%%              after computing the unvariance of the reconstruction %%%%%%
%%              error, which are: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% - X_rec:     Final scaled data that can be reconstructed for the given %
%%              set of faults and sensors in the statistical model. %%%%%%%
%% - l_rec:     Number of principal components that have an %%%%%%%%%%%%%%%
%%              unreconstructed variance lower than the desired %%%%%%%%%%%
%%              threshold. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% - xi_rec:    Set of faults that can be correctly detected, %%%%%%%%%%%%%
%%              reconstructed and diagnosed using l_rec number of %%%%%%%%%
%%              principal components. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% - U_local:	Matrix with the unreconstructed variance for each fault %%%
%%              in xi_rec and every possible number of principal %%%%%%%%%%
%%              components. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% - U_global:	Vector with the unreconstructed variance for all the %%%%%%
%%              faults at the same time with every possible number of %%%%%
%%              principal components. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% - sensors:	Vector that contains those sensors that can be %%%%%%%%%%%%
%%              reconstruted at each iteration. %%%%%%%%%%%%%%%%%%%%%%%%%%%
%% - faults:	Vector containing the indexs of those faults that can be %%
%%              reconstructed. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% - factors:	Vector containing the factors used to decide whether a %%%%
%%              fault can be reconstruced or not. %%%%%%%%%%%%%%%%%%%%%%%%%
%% - xi_pcs:	Projection of the matrix xi in the principal components %%%
%%              subspace. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% - xi_res:	Porjection of the matrix xi in the residual subspace. %%%%%
%%              This matrix is required for fault diagnosis and %%%%%%%%%%%
%%              reconstruction. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% - n_loops:   Number of iterations the method required to find the %%%%%%
%%              reconstructible sensors and the number of principal %%%%%%%
%%              componentes to use. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function VRE_Struct = variance_reconstruction_error(X, xi)
% We initialise the output parameter to the worst case. In this case, the
% worst situation would be that we remove only one sensor at a time, so we
% will define the size of the output parameters to the number of sensors at
% the first iteration.
% - Data set to be used for the model generation
VRE_Struct.X_rec = cell(1,size(X,2));
% - Number of principal components that present that satisfies the
% variance of reconstruction error condition (minimum or at least
% rec_tolerance)
VRE_Struct.l_rec = cell(1,size(X,2));
% - Matrix relating faults and sensors faults that can be reconstructed
% with the variance of reconstruction error condition
VRE_Struct.xi_rec = cell(1,size(X,2));
% - Matrix containing the relation of faults and the variance of
% reconstruction error for every possible number of principal components
VRE_Struct.U_global = cell(1,size(X,2));
% - Vector containing the relation of the variance of reconstruction error
% for all the matrix faults that will be considered in order to find the
% global satisfaction condition optimal
VRE_Struct.U_local = cell(1,size(X,2));
% - Vector containing all the sensors that satisfy the variance of
% reconstruction error condition
VRE_Struct.sensors = cell(1,size(X,2));
% - Vector containing all the faults that satisfy the variance of
% reconstruction error condition
VRE_Struct.faults = cell(1,size(xi,2));
% - Vector containing the factors used to decide whether a fault can be
% reconstructed or not
VRE_Struct.factors = cell(1,size(X,2));
% - Vector containing the projection of the fault matrix xi in the
% principal components space
VRE_Struct.xi_pcs = cell(1,size(X,2));
% - Vector containing the projection of the fault matrix in the residual
% subspace
VRE_Struct.xi_res = cell(1,size(X,2));
% By default we initialise the loop control variable to 1
loop = 1;
% We set the parameters to compute all the variables to the original ones
X_loop = X;
xi_loop = xi;
sensors_loop = 1:size(xi,1);
faults_loop = 1:size(xi,2);
% We initialise the position to write the results of iteration i
j = 1;
% While there are changes...
while loop == 1
    % We compute the covariance matrix
    R_loop = (X_loop'*X_loop) / (size(X_loop, 1) - 1);
    % We initialise the reconstruction error matrix for this loop
    U_local_loop = zeros(size(xi_loop,2),size(X_loop, 2));
    % We compute the full reconstruction error matrix for each fault and number
    % of principal components
    for i = 1:size(xi_loop,2)
        [U_local_loop(i,:),~,~] = compute_reconstruction_error_matrix(X_loop,xi_loop(:,i));
    end
    % We compute the reconstruction error matrix for the whole fault set
    [U_global_loop,xi_pcs_loop,xi_res_loop] = compute_reconstruction_error_matrix(X_loop,xi_loop);
    % We find the minimum value of unreconstructed variance in the whole data set that
    % will be the optimal of this loop
    [~,l_rec_loop] = min(U_global_loop(1:length(U_global_loop) - 1));
    % We compute the factor to decide wheter a fault can be reconstructed
    % enough
    factor = diag(xi_loop' * R_loop * xi_loop);
    % We give value to the output parameters
    VRE_Struct.X_rec{j} = X_loop;
    VRE_Struct.l_rec{j} = l_rec_loop;
    VRE_Struct.xi_rec{j} = xi_loop;
    VRE_Struct.sensors{j} = sensors_loop;
    VRE_Struct.U_local{j} = U_local_loop;
    VRE_Struct.U_global{j} = U_global_loop;
    VRE_Struct.faults{j} = faults_loop;
    VRE_Struct.factors{j} = factor;
    VRE_Struct.xi_pcs{j} = xi_pcs_loop;
    VRE_Struct.xi_res{j} = xi_res_loop;
    % With the l optimal number of principal components we find if there
    % are any faults that cannot be reconstructed using the l_opt_rec_loop
    % number of principal components
    reconstructibility = U_local_loop(:,l_rec_loop) > factor;
    % We find those indices that are greater than the factor
    % (non_reconstructible faults)
    non_reconstructible = find(reconstructibility);
    % We update the loop control variable
    loop = ~isempty(non_reconstructible) & length(sensors_loop) ~= ...
        length(non_reconstructible);
    % If there are non reconstructible faults we have to update the
    % parameters for the next loop, otherwise, the algorithm ends here
    if loop == 1
        % We find the sensors associated to the non reconstructable faults
        [rows,~] = find(xi_loop(:,non_reconstructible));
        % We erase those faults that cannot be reconstructed from the fault
        % set
        xi_loop(:,non_reconstructible) = [];
        % We remove from the fault data set those that are not
        % reconstructible
        faults_loop(:,non_reconstructible) = [];
        % We erase those sensors associated to non reconstructible sensors
        xi_loop(rows,:) = [];
        % We eliminate the non reconstructible faults associated sensors
        % from the data set
        X_loop(:,rows) = [];
        % We remove those sensors that cannot be reconstructed
        sensors_loop(rows) = [];
        % We increment the position to write the next result
        j = j + 1;
    end
end
% We look if there were any sensor that was reconstructible at the end
if length(sensors_loop) == length(non_reconstructible)
    % We advise the user that all sensors would be removed that there are 
    % no sensors to be reconstructed
    fprintf('At iteration %d all sensors cannot be reconstruced!!\n',j);
end
% Finally, we only have to remove those positions of each variable that has
% not been used
VRE_Struct.X_rec(j+1:size(X,2)) = [];
VRE_Struct.l_rec(j+1:size(X,2)) = [];
VRE_Struct.xi_rec(j+1:size(X,2)) = [];
VRE_Struct.sensors(j+1:size(X,2)) = [];
VRE_Struct.U_local(j+1:size(X,2)) = [];
VRE_Struct.U_global(j+1:size(X,2)) = [];
VRE_Struct.faults(j+1:size(xi,2)) = [];
VRE_Struct.factors(j+1:size(X,2)) = [];
VRE_Struct.xi_pcs(j+1:size(X,2)) = [];
VRE_Struct.xi_res(j+1:size(X,2)) = [];
VRE_Struct.n_loops = j;